import argparse
import datetime
import os

import arakawa as ar  # type: ignore
import numpy as np
import pandas as pd
import plotly.express as px  # type: ignore

import ncbi_cluster_tracker.cluster as cluster

from ncbi_cluster_tracker.logger import logger

class ClusterReport:
    """
    Contains Arakawa Report for "Cluster details" tab with SNP cluster metrics,
    distance matrix heatmap, and isolate count over time.
    """
    PRIMARY_METADATA_COLS = [
        'isolate_id',
        'collection_date',
        'geo_loc_name',
    ]

    def __init__(
        self, 
        cluster: cluster.Cluster,
        metadata: pd.DataFrame,
        clusters_df: pd.DataFrame,
    ):

        self.cluster = cluster
        self.clusters_df = clusters_df
        self.metadata = metadata
        self.metadata_truncated = self._truncate_metadata(metadata.copy()).fillna('')
        self.snp_matrix = self._create_snp_matrix()
        self.custom_labels = self._create_custom_labels()
        self.report = self._create_report()

    def _create_snp_matrix(self) -> ar.Group:
        """
        Create heatmap of SNP distance matrix annotated with metadata.
        """
        if self.cluster.filtered_matrix is None:
            text = f'⚠️ WARNING: {self.cluster.filtered_matrix_message}'
            snp_matrix = ar.Group(
                ar.Text(text, name=self.cluster.name, label=self.cluster.name),
            )
            return snp_matrix

        matrix = self._add_metadata_to_matrix(self.cluster.filtered_matrix)
        matrix = self._rename_matrix_ids(matrix)
        # Some isolates may be duplicated due to multiple assemblies, keep first
        matrix = matrix[~matrix.index.duplicated(keep='first')]
        matrix = matrix.loc[:, ~matrix.columns.duplicated(keep='first')]

        style = (matrix
            .style
            .background_gradient()
            .set_table_styles([
                {
                    'selector': 'th',
                    'props': [
                        ('width', '40px'),
                        ('padding', '3px'),
                        ('font-size', '12px')
                    ]
                },
                {
                    'selector': 'th.col_heading',
                    'props': [
                        ('writing-mode', 'vertical-rl'),
                        ('transform', 'rotateZ(180deg)'), 
                        ('height', '120px'),
                        ('vertical-align', 'middle'),
                        ('horizontal-align', 'left'),
                    ]
                },
                # {
                #     'selector': 'thead th.blank',
                #     'props': [('display', 'none')]
                # },
                # {
                #     'selector': 'thead th.index_name',
                #     'props': [('display', 'none')]
                # },
                {
                    'selector': 'td',
                    'props': [
                        ('font-size', '12px'),
                        ('padding', '3px'),
                        ('white-space', 'nowrap'),
                        ('overflow', 'hidden'),
                        ('text-overflow', 'ellipsis'),
                    ]
                }
            ])
            .set_sticky(axis=0)
        )
        if self.cluster.filtered_matrix_message is not None:
            text = f'⚠️ WARNING: {self.cluster.filtered_matrix_message}'
            snp_matrix = ar.Group(
                ar.Text(text, name=self.cluster.name, label=self.cluster.name),
                ar.Table(style, label=self.cluster.name)
            )
        else:
            snp_matrix = ar.Group(ar.Table(style, label=self.cluster.name))
        return snp_matrix

    def _rename_matrix_ids(self, matrix: pd.DataFrame) -> pd.DataFrame:
        """
        Use 'id' column from sample sheet to name internal isolates in the
        matrix and also mark them with a star. Use BioSample ID for external
        isolates and for internal isolates if 'id' column not provided.
        """
        self.metadata = self.metadata.set_index('target_acc')

        if 'id' in self.metadata.columns:
            internal = self.metadata.query('source == "internal"').copy()
            internal['label'] = internal['id']
        else:
            internal = self.metadata.query('source == "internal"').copy()
            internal['label'] = internal['biosample']

        external = self.metadata.query('source == "external"').copy()
        external['label'] = external['biosample']

        def apply_prefix(row: pd.Series) -> str:
            if row.get('is_new') == 'yes':
                if row.get('source') == 'internal':
                    return f'⭐🆕{row.label}'
                return f'🆕{row.label}'
            elif row.get('source') == 'internal':
                return f'⭐{row.label}'
            return str(row.label)

        combined = pd.concat([
            internal,
            external if not external.empty else None,
        ])
        combined['renamed'] = combined.apply(apply_prefix, axis=1)

        renamer = combined['renamed']
        renamer.index = combined.index  # ensure target_acc is the index
        matrix = matrix.rename(columns=renamer.to_dict(), index=renamer.to_dict())

        self.metadata = self.metadata.reset_index()
        return matrix
    
    def _add_metadata_to_matrix(self, matrix: pd.DataFrame) -> pd.DataFrame:
        """
        Add columns with basic metadata to the distance matrix.
        """
        columns = self.PRIMARY_METADATA_COLS.copy()
        if 'filtered_amr' in self.metadata.columns:
            columns.append('filtered_amr')
        matrix = (self.metadata_truncated[['target_acc', *columns]]
            .set_index('target_acc')
            .merge(
                matrix,
                right_index=True,
                left_on='target_acc',
                how='right',
            )
        )
        matrix.index.name = ''
        return matrix

    def _truncate_metadata(self, metadata: pd.DataFrame) -> pd.DataFrame:
        """
        Limit string length displayed in report to prevent text from wrapping
        and wide column widths.
        """
        def truncate(text: str) -> str:
            if isinstance(text, str) and len(text) > 15:
                return f'{text[:7]}...{text[-7:]}'
            return text

        metadata[self.PRIMARY_METADATA_COLS] = (
            metadata[self.PRIMARY_METADATA_COLS].map(lambda x: truncate(x))
        )
        return metadata 

    def _count_isolates(self) -> tuple[ar.Group, ar.Text | None]:
        """
        Summarize internal and external isolate counts.
        """
        internal_change: np.int64 | None
        external_change: np.int64 | None
        total_change: np.int64 | None
        internal_count = np.int64(self.clusters_df['internal_count'].iloc[0].item())
        external_count = np.int64(self.clusters_df['external_count'].iloc[0].item())
        total_count = internal_count + external_count
        internal_change = np.int64(self.clusters_df['internal_change'].iloc[0].item())
        external_change = np.int64(self.clusters_df['external_change'].iloc[0].item())
        total_change = internal_change + external_change

        if internal_change > 0:
            internal_upward = True
        elif internal_change < 0:
            internal_upward = False
        else:
            internal_change = None
            internal_upward = None

        if external_change > 0:
            external_upward = True
        elif external_change < 0:
            external_upward = False
        else:
            external_change = None
            external_upward = None

        if total_change > 0:
            total_upward = True
        elif total_change < 0:
            total_upward = False
        else:
            total_change = None
            total_upward = None

        # Isolate counts may differ between FTP and BigQuery datasets
        # if the FTP site updated before BigQuery or the FTP downloaded data is
        # outdated (second case less likely). In either case the user should be
        # warned of the mismatch.
        warning_message = []
        if self.cluster.external_isolates is not None:
            external_count_ftp = len(self.cluster.external_isolates)
        else:
            external_count_ftp = 0

        if external_count != external_count_ftp:
            warning_message = [ar.Text(
                '⚠️ WARNING: A more up-to-date version of this cluster may be ' \
                'available on the Pathogen Detection site with more ' \
                'internal and/or external isolates (visit the backup link).' \
            )]

        # List specific new isolates
        MAX_DISPLAY = 5
        if internal_change is not None:
            if 'id' in self.metadata.columns:
                internal_new = self.metadata.query(
                    'source == "internal" and is_new == "yes"'
                )[['id', 'isolate_id']]
            else:
                internal_new = self.metadata.query(
                    'source == "internal" and is_new == "yes"'
                )[['biosample', 'isolate_id']]
            internal_news = internal_new.iloc[:, 0].str.cat(internal_new.iloc[:, 1], sep=' / ').tolist()
            internal_list = 'New internal isolates added:\n'
            internal_list = f"{internal_list} - {'\n - '.join(internal_news[:MAX_DISPLAY])}"
            if len(internal_news) > MAX_DISPLAY:
                internal_list = f'{internal_list}\nand {internal_change - MAX_DISPLAY} mor.'
        else:
            internal_list = 'No change in internal isolate count.'

        if external_change is not None:
            external_new = self.metadata.query(
                'source == "external" and is_new == "yes"'
            )[['biosample', 'isolate_id']]
            external_news = external_new.iloc[:, 0].str.cat(external_new.iloc[:, 1], sep=' / ').tolist()
            external_list = 'New external isolates added:\n'
            external_list = f"{external_list} - {'\n - '.join(external_news[:MAX_DISPLAY])}\n"
            if len(external_news) > MAX_DISPLAY:
                external_list = f'{external_list}- and {external_change - MAX_DISPLAY} more'
        else:
            external_list = 'No change in external isolate count.'

        count_blocks: tuple[ar.Group, ar.Text | None] = ar.Group(
            blocks=[
                *warning_message,
                ar.Group(
                    ar.BigNumber(
                        heading='Internal isolates',
                        value=internal_count,
                        change=internal_change,
                        is_upward_change=internal_upward,
                    ),
                    ar.BigNumber(
                        heading='External isolates',
                        value=external_count,
                        change=external_change,
                        is_upward_change=external_upward,
                    ),
                    ar.BigNumber(
                        heading='Total isolates',
                        value=total_count,
                        change=total_change,
                        is_upward_change=total_upward,
                    ),
                    ar.Text(internal_list),
                    ar.Text(external_list),
                    columns=3,
                ),
            ]
        )
        return count_blocks

    def _create_isolate_count_graph(self) -> ar.Plot:
        """
        Histogram of isolate counts over creation date (similar to 'epi curve').
        """
        # TODO: show sample ID on hover
        # TODO: collection dates, date added, or both?
        # If isolate only has year or no collection date, use date_added,
        # otherwise, use the collection date?
        count_graph = ar.Plot(
            px.histogram(
                self.metadata,
                x='creation_date',
                color='source',
            )
        )
        return count_graph

    def _create_custom_labels(self) -> ar.Attachment:
        """
        Create custom labels file that can be uploaded to the NCBI Pathogen
        Detection SNP tree viewer
        """
        cols = self.PRIMARY_METADATA_COLS.copy()
        if 'id' in self.metadata.columns:
            cols.insert(0, 'id')
        if 'filtered_amr' in self.metadata.columns:
            cols.append('filtered_amr')

        # prefix label with star to avoid collision with Pathogen Detection
        star_cols = {k: f'*{k}' for k in cols}

        custom_labels = (
            self.metadata
            .rename(columns=star_cols)
            .melt(
                id_vars=['target_acc'],
                value_vars=list(star_cols.values()),
            )
        )
        subdir = os.path.join(os.environ['NCT_OUT_SUBDIR'], 'labels')
        os.makedirs(subdir, exist_ok=True)
        path = os.path.join(subdir, f'{self.cluster.name}_labels.txt')
        custom_labels.to_csv(path, sep='\t', index=False, header=False) 
        attachment = ar.Attachment(file=path)
        return attachment 

    def _create_report(self) -> ar.Group:
        """
        Combine tables and visualizations for the cluster into an Arawaka
        Group to be displayed together in the report.
        """
        taxgroup_name = self.clusters_df[
            self.clusters_df['cluster'] == self.cluster.name
        ]['taxgroup_name'].item()
        title = ar.HTML(f'<h2><i>{taxgroup_name}</i> cluster {self.cluster.name}</h2>')
        count_blocks = self._count_isolates()
        tree_header = ar.HTML('<h3>NCBI Pathogen Detection</h3>')
        tree_url = self.clusters_df[
            self.clusters_df['cluster'] == self.cluster.name
        ]['tree_url'].item()
        browser_link_base = 'https://www.ncbi.nlm.nih.gov/pathogens/isolates/#'

        # Modify URL so that internal isolates are selected red
        new_internals = self.metadata.query('source == "internal" and is_new == "yes"')['target_acc'].tolist()
        if new_internals:
            tree_url_all_internal = f'{tree_url}?accessions={','.join(self.cluster.internal_isolates)}&accessions2={','.join(new_internals)}'
        else:
            tree_url_all_internal = f'{tree_url}?accessions={','.join(self.cluster.internal_isolates)}'
        backup_link_all_internal = f'{browser_link_base}{'%20'.join(self.cluster.internal_isolates)}'
        tree_links = (
            f'Links to tree:\n\n' \
            f'- [Select all internal isolates (bold new internal)]({tree_url_all_internal}) ([Backup link]({backup_link_all_internal}))\n'
        )
        if new_internals:
            tree_url_new_internal = f'{tree_url}?accessions={','.join(new_internals)}&accessions2={','.join(self.cluster.internal_isolates)}'
            backup_link_new_internal = f'{browser_link_base}{'%20'.join(new_internals)}'
            tree_links = f'{tree_links}- [Select new internal isolates (bold all internal)]({tree_url_new_internal}) ([Backup link]({backup_link_new_internal}))\n'
        new_all = self.metadata.query('is_new == "yes"')['target_acc'].tolist()
        if new_all:
            tree_url_new_all = f'{tree_url}?accessions={','.join(new_all)}&accessions2={','.join(self.cluster.internal_isolates)}'
            backup_link_new_internal = f'{browser_link_base}{"%20".join(new_all)}'
            tree_links = f'{tree_links}- [Select all new isolates (bold all internal)]({tree_url_new_all}) ([Backup link]({backup_link_new_internal}))\n'
        cluster_base = self.cluster.name.split('.')[0]
        tree_links_block = ar.Text(tree_links)

        snp_header = ar.HTML('<h3>SNP distance matrix</h3>')
        count_graph_header = ar.HTML('<h3>Isolates by date added</h3>')
        count_graph = self._create_isolate_count_graph()
        table_header = ar.HTML('<h3>Isolate table</h3>')
        table = ar.DataTable(
            self.metadata
            .sort_values(by=['source', 'creation_date'], ascending=[False, False])
            .reset_index()
            .drop(columns='index')
        )
        if (
            self.clusters_df['change'].iloc[0] == 'new cluster'
            or not self.clusters_df['change'].iloc[0].startswith('+0')
        ):
            new = '🆕'
        else:
            new = ''
        blocks = [
                title,
                count_blocks,
                tree_header,
                tree_links_block,
                self.custom_labels,
                snp_header,
                self.snp_matrix,
                count_graph_header,
                count_graph,
                table_header,
                table,
        ]
        report = ar.Group(
            blocks = [b for b in blocks if b is not None],
            label=f'{new} {self.cluster.name} - {taxgroup_name}'
        )
        return report


def mark_new_isolates(
    isolates_df: pd.DataFrame,
    old_isolates_df: pd.DataFrame | None,
) -> pd.DataFrame:
    """
    Add "is_new" column to `isolates_df` to indicate any new isolates added
    compared to `old_cluster_df`.
    """
    if old_isolates_df is None:
        isolates_df['is_new'] = 'yes'
        return isolates_df
    old_biosamples = set(old_isolates_df['biosample'])
    isolates_df['is_new'] = isolates_df['biosample'].apply(
        lambda x: 'yes' if x not in old_biosamples else 'no'
    )
    return isolates_df


def combine_metadata(
    sample_sheet_df: pd.DataFrame,
    isolates_df: pd.DataFrame,
    amr_df: pd.DataFrame | None,
    args: argparse.Namespace,
) -> pd.DataFrame:
    """
    Combine user-provided data in sample sheet with sample metadata
    from NCBI.
    """
    metadata = sample_sheet_df.merge(
        isolates_df,
        how='outer',
        on='biosample',
        suffixes=('__internal', '__external'),
        indicator='source'
    )
    # Use sample sheet values for matching columns in sample sheet and BigQuery
    internal_merged_cols = [c for c in metadata.columns if '__internal' in c]
    for internal_col in internal_merged_cols:
        base_col = internal_col[:-10]  # without __suffix
        external_col = f'{base_col}__external'
        metadata[base_col] = np.where(
            ~metadata[internal_col].isnull(),
            metadata[internal_col],
            metadata[external_col],
        )
        metadata = metadata.drop([internal_col, external_col], axis=1)

    metadata['source'] = metadata['source'].map({
        'left_only': 'internal',
        'right_only': 'external',
        'both': 'internal',
    })
    optional_id = ['id'] if 'id' in metadata.columns else []

    filtered_amr_col = []
    if amr_df is not None and args.filter_amr:
        reduced_df = amr_df[['biosample', 'element']]
        reduced_df = reduced_df.groupby('biosample')['element'].agg(','.join).reset_index()
        reduced_df = reduced_df.rename(columns={'element': 'filtered_amr'})
        metadata = pd.merge(metadata, reduced_df, how='left', on='biosample')
        filtered_amr_col = ['filtered_amr']

    column_order = [
        'biosample',
        *optional_id,
        'isolate_id',
        'cluster',
        'source',
        'is_new',
        'scientific_name',
        'collection_date',
        'creation_date',
        'geo_loc_name',
        'isolation_source',
        'taxgroup_name',
        'bioproject_acc',
        'target_acc',
        'sra_id',
        *filtered_amr_col,
    ]
    extra_internal_cols = [c for c in metadata.columns if c not in column_order]
    extra_internal_cols.sort()
    column_order.extend(extra_internal_cols)
    metadata = metadata[column_order]
    return metadata


def create_cluster_reports(
    clusters: list[cluster.Cluster],
    clusters_df: pd.DataFrame,
    metadata: pd.DataFrame
) -> list[ClusterReport]:
    """
    Create a ClusterReport for all clusters.
    """
    cluster_reports: list[ClusterReport] = []
    for cluster in clusters:
        cluster_report = ClusterReport(
            cluster,
            metadata[metadata['cluster'] == cluster.name],
            clusters_df[clusters_df['cluster'] == cluster.name],
        )
        cluster_reports.append(cluster_report)
    return cluster_reports


def add_counts(clusters_df: pd.DataFrame, metadata: pd.DataFrame) -> pd.DataFrame:
    """
    Add external and total isolate counts to `clusters_df` DataFrame.
    """
    if 'total_count' not in clusters_df.columns:
        clusters_df['total_count'] = (
            clusters_df['internal_count'] + clusters_df['external_count']
        )
    else:
        internal_counts = (metadata
            .query('source == "internal"')
            .loc[:, 'cluster']
            .value_counts()
            .rename('internal_count')
        )
        clusters_df = clusters_df.merge(internal_counts, on='cluster')
        clusters_df['external_count'] =  (
            clusters_df['total_count'] - clusters_df['internal_count']
        )
    return clusters_df


def compare_counts(
    clusters_df: pd.DataFrame,
    old_clusters_df: pd.DataFrame | None,
) -> pd.DataFrame:
    """
    Compare cluster isolate counts between previous report, if provided, and add
    differences to the `clusters_df` DataFrame.
    """
    clusters_df['cluster_base'] = clusters_df['cluster'].str.split('.').str[0]
    clusters_df = clusters_df.astype({'internal_count': 'Int64', 'external_count': 'Int64'})

    if old_clusters_df is None:
        clusters_df['internal_change'] = 0
        clusters_df['external_change'] = 0
        clusters_df['change'] = 'new cluster'
        return clusters_df

    old_clusters_df['cluster_base'] = old_clusters_df['cluster'].str.split('.').str[0]
    old_clusters_df = old_clusters_df.astype({'internal_count': 'Int64', 'external_count': 'Int64'})

    compare_df = pd.merge(
        clusters_df[['cluster_base', 'internal_count', 'external_count']],
        old_clusters_df[['cluster_base', 'internal_count', 'external_count']],
        on='cluster_base',
        how='left',
        suffixes=('_new', '_old'),
    )
    compare_df['internal_change'] = compare_df['internal_count_new'] - compare_df['internal_count_old']
    compare_df['external_change'] = compare_df['external_count_new'] - compare_df['external_count_old']

    def create_change_column(row: pd.Series) -> str:
        if pd.isna(row['internal_change']) or pd.isna(row['external_change']):
            return 'new cluster'
        internal_prefix = '+' if row['internal_change'] >= 0 else ''
        external_prefix = '+' if row['external_change'] >= 0 else ''
        return f'{internal_prefix}{row["internal_change"]} / {external_prefix}{row["external_change"]}'

    compare_df['change'] = compare_df.apply(
        create_change_column,
        axis=1
    )
    if 'change' in clusters_df.columns:
        clusters_df = clusters_df.drop(columns='change')

    clusters_df = pd.merge(
        compare_df[['cluster_base', 'change', 'internal_change', 'external_change']],
        clusters_df,
        how='right',
        on='cluster_base',
    )
    clusters_df['internal_change'] = clusters_df['internal_change'].fillna(0)
    clusters_df['external_change'] = clusters_df['external_change'].fillna(0)
    return clusters_df

def create_clusters_timeline_plot(
    metadata: pd.DataFrame,
    previous_max_date: datetime.datetime | None
) -> ar.Plot:
    """
    Create plot showing when each isolate was added to each cluster. Add
    vertical red line to plot showing when the previous report's last
    isolate was
    """
    MAX_DISPLAY = 15

    # px.strip() wasn't jittering points with datetime, so manually jittering
    metadata_jittered = metadata.copy()
    metadata_jittered = metadata_jittered[metadata_jittered['cluster'].notna()]
    metadata_jittered = metadata_jittered.sort_values(by='creation_date', ascending=False)
    metadata_jittered['cluster_ticktext'] = (
        metadata_jittered['cluster'].str.cat(metadata_jittered['taxgroup_name'], sep='<br>')
    )
    message = None
    if metadata_jittered['cluster'].nunique() > MAX_DISPLAY:
        top_clusters = metadata_jittered['cluster'].unique()[:MAX_DISPLAY]
        metadata_jittered = metadata_jittered[metadata_jittered['cluster'].isin(top_clusters)]
        message = f'Top {MAX_DISPLAY} most recently updated clusters displayed'
    metadata_jittered['cluster_jittered'] = (
        pd.Categorical(metadata_jittered['cluster'], categories=metadata_jittered['cluster'].unique(), ordered=True).codes
        + np.random.uniform(-0.10, 0.10, size=len(metadata_jittered))
    )
    fig = px.scatter(
        metadata_jittered,
        x='creation_date',
        y='cluster_jittered',
        color='source',
        height=90 * len(metadata_jittered['cluster'].unique()),
        hover_data=['cluster', 'isolate_id'],
    ).update_yaxes(
        tickmode='array',
        tickvals=list(range(len(metadata_jittered['cluster'].unique()))),
        ticktext=metadata_jittered['cluster_ticktext'].unique(),
        autorange='reversed',
    ).update_traces(
        hovertemplate='<b>Isolate ID:</b> %{customdata[1]}<br><b>Date:</b> %{x}'
    ).update_layout(
        yaxis_title='cluster',
        xaxis={'side': 'top'},
    )
    if previous_max_date is not None:
        fig.add_vline(
            x=previous_max_date.timestamp() * 1000,  # https://github.com/plotly/plotly.py/issues/3065
            line_dash='dash',
            line_color='indianred',
            annotation_text='previous report'
        )
    plot = ar.Plot(fig)
    return plot, message


def write_final_report(
     clusters_df: pd.DataFrame,
     old_clusters_df: pd.DataFrame | None,
     clusters: list[cluster.Cluster], 
     metadata: pd.DataFrame,
     amr_df: pd.DataFrame | None,
     compare_dir: str | None,
     command: str
) -> None:
    """
    Output final, standalone HTML report with all tables and plots. This
    function also outputs the clusters CSV.
    """
    logger.info('Generating HTML report...')
    clusters_df = add_counts(clusters_df, metadata)
    clusters_df = compare_counts(clusters_df, old_clusters_df)
    cluster_reports = create_cluster_reports(
        clusters,
        clusters_df,
        metadata
    )
    keep_cols = [
        'cluster',
        'taxgroup_name',
        'internal_count',
        'external_count',
        'change',
        'latest_added',
        'earliest_added',
        'earliest_year_collected',
        'latest_year_collected',
        'tree_url',
    ]
    clusters_df = clusters_df[keep_cols]
    clusters_csv = os.path.join(
        os.environ['NCT_OUT_SUBDIR'],
        f'clusters_{os.environ["NCT_NOW"]}.csv'
    )
    clusters_df.to_csv(clusters_csv, index=False)

    cluster_page_blocks = [ar.HTML(f'<h2>Cluster report {os.environ["NCT_NOW"]}</h2>')]
    command_header = ar.Text('Command: ')
    command_block = ar.Code(code=command, language='javascript')
    cluster_page_blocks.extend([command_header, command_block])
    isolate_page_blocks = []
    if compare_dir is not None:
        header_2 = ar.Text(f'↔️ Comparing to {compare_dir}')
        cluster_page_blocks.append(header_2)
        isolate_page_blocks.append(header_2)

    clusters_table = ar.DataTable(
        clusters_df
            .sort_values(['change', 'latest_added'], ascending=[False, False])
            .reset_index()
            .drop(columns='index')
    )
    previous_max_date = (
        pd.to_datetime(old_clusters_df['latest_added'].max()) if old_clusters_df is not None else None
    )
    clusters_timeline_plot, clusters_timeline_message = create_clusters_timeline_plot(
        metadata,
        previous_max_date,
    )
    cluster_page_blocks.extend([
        clusters_table,
        ar.HTML('<h3>Cluster timelines</h3>'),
    ])
    if clusters_timeline_message is not None: 
        cluster_page_blocks.append(clusters_timeline_message)
    cluster_page_blocks.append(clusters_timeline_plot)

    isolates_table = ar.DataTable(
        metadata
        .sort_values(by=['source', 'creation_date'], ascending=[False, False])
        .reset_index()
        .drop(columns='index')
    )
    isolate_page_blocks.append(isolates_table)
    missing_isolates = metadata.query('source == "internal" and is_new.isna()')['biosample'].tolist()
    missing_isolates.sort(reverse=True)
    metadata['is_new'] = metadata['is_new'].fillna('(no data)')
    # TODO: List specific reason for missing data, at least when possible
    if missing_isolates:
        missing_message = ar.Text(
            f'⚠️ WARNING: No cluster data displayed for the internal '\
            f'isolates listed below, either because they have not finished '\
            f'processing, failed quality control, or are otherwise not '\
            f'included in the system. More information about these isolates '\
            f'may be available in the Isolates Browser search results '\
            f'[here](https://www.ncbi.nlm.nih.gov/pathogens/isolates/#{"%20".join(missing_isolates)})'
            f'\n\n{", ".join(missing_isolates)}'
        )
        isolate_page_blocks.append(missing_message)

    cluster_report_blocks = [r.report for r in cluster_reports]
    cluster_report_blocks.sort(key=lambda r: r.label, reverse=True)

    report_blocks = [
        ar.Page(blocks=cluster_page_blocks, title='Clusters'),
        ar.Page(
            ar.Select(blocks=cluster_report_blocks, type=ar.SelectType.DROPDOWN),
            title='Cluster details',
        ),
        ar.Page(blocks=isolate_page_blocks, title='Isolates'),
    ]

    if amr_df is not None:
        amr_blocks = []
        if 'filtered_amr' in metadata:
            args = command.split(' ')  
            for i in range(len(args)):
                if args[i] == '--filter-amr':
                    filters = args[i + 1]
            filtered_message = ar.Text(f'Table filtered to only show the following CLASS:SUBCLASS pairs: {filters}')
            amr_blocks.append(filtered_message)
        amr_table = ar.DataTable(
            amr_df
            .sort_values('biosample', ascending=False)
            .reset_index()
            .drop(columns='index')
        )
        amr_blocks.append(amr_table)
        amr_page = ar.Page(blocks=amr_blocks, title='AMR')
        report_blocks.append(amr_page)

    report = ar.Blocks(blocks=report_blocks)

    ar.save_report(
        report,
        path=os.path.join(
            os.environ['NCT_OUT_SUBDIR'],
            f'clusters_{os.environ['NCT_NOW']}.html'
        ),
        standalone=False,
    )



import io
import os
import re
import requests
import shutil

import numpy as np
import pandas as pd
import tqdm

from ncbi_cluster_tracker.logger import logger

# Can't auto-translate these to folders on NCBI FTP
TAXGROUP_TO_ORGANISM = {
    'Acinetobacter baumannii': 'Acinetobacter',
    'Aeromonas hydrophila': 'Aeromonas',
    'Campylobacter jejuni': 'Campylobacter',
    'E.coli and Shigella': 'Escherichia_coli_Shigella',
    'Elizabethkingia anophelis': 'Elizabethkingia',
    'Klebsiella pneumoniae': 'Klebsiella',
    'Listeria monocytogenes': 'Listeria',
    'Morganella morganii': 'Morganella',
    'Providencia alcalifaciens': 'Providencia',
    'Salmonella enterica': 'Salmonella',
    'Serratia marcescens': 'Serratia',
}


def download_cluster_files(clusters_df: pd.DataFrame, keep_files: bool) -> None:
    clusters_df['snp_url'] = clusters_df.apply(build_snp_url, axis=1)
    urls = clusters_df['snp_url'].to_list()
    download_snps(urls, keep_files)


def build_ftp_base_url(taxgroup_name: str) -> str:
    url = 'https://ftp.ncbi.nlm.nih.gov/pathogen/Results/'
    url += get_organism_from_taxgroup(taxgroup_name)
    url += '/latest_snps/'
    return url


def get_organism_from_taxgroup(taxgroup_name: str) -> str:
    if taxgroup_name in TAXGROUP_TO_ORGANISM:
        organism = f'{TAXGROUP_TO_ORGANISM[taxgroup_name]}'
    else:
        organism = f'{taxgroup_name.replace(' ', '_')}'
    return organism


def build_snp_url(row: pd.Series) -> str:
    url = build_ftp_base_url(row['taxgroup_name'])
    url += f'SNP_trees/{row['cluster']}.tar.gz'
    return url


def get_latest_kmer_group_acc(taxgroup_name: str) -> str:
    # BigQuery doesn't give us the specific kmer_group_acc, so we'll have to 
    # pull it from the XML descriptor file on the FTP
    url = build_ftp_base_url(taxgroup_name)
    response = requests.get(url)
    if not response.ok:
        raise Exception('Could not find k-mer group accession: {url}')
    acc_match = re.search(r'PDG(\d+)\.(\d+)', response.text)
    if acc_match is not None:
        acc = acc_match.group(0) 
    else:
        raise Exception(f'Could not find k-mer group accession: {url}')
    return acc


def build_tree_viewer_url(
        row: pd.Series,
        taxgroup_to_kmer_group_acc: dict[str, str] = {}
) -> str:
    taxgroup_name = row['taxgroup_name']
    if row['cluster'] not in taxgroup_to_kmer_group_acc:
        taxgroup_to_kmer_group_acc[taxgroup_name] = (
            get_latest_kmer_group_acc(taxgroup_name)
        )
    url = 'https://www.ncbi.nlm.nih.gov/pathogens/tree#'
    url += get_organism_from_taxgroup(taxgroup_name)
    url += '/'
    url += taxgroup_to_kmer_group_acc[taxgroup_name]
    url += '/'
    url += str(row['cluster'])
    return url


def download_snps(urls: list[str], keep_files: bool) -> None:
    out_subdir = os.path.join(os.environ['NCT_OUT_SUBDIR'], 'snps')
    logger.info(f'Downloading SNP cluster data to {out_subdir}...')
    os.makedirs(out_subdir, exist_ok=True)
    for url in tqdm.tqdm(urls):
        with requests.Session() as session:
            for _ in range(3):
                response = session.get(url)
                if response.ok:
                    break
                # try incrementing cluster version
                regex = r'(?<=\.)(\d+)(?=\.tar\.gz)'
                url = re.sub(regex, lambda x: str(int(x.group())+1), url)
            else:
                raise Exception(f'Could not find cluster: {url}')

            destination = os.path.join(out_subdir, os.path.basename(url))
            with open(destination, 'wb') as f:
                f.write(response.content)
            shutil.unpack_archive(destination, out_subdir)
            out_files = os.listdir(out_subdir)
            for out_file in out_files:
                if not out_file.endswith('.newick') and not keep_files:
                    os.remove(os.path.join(out_subdir, out_file))
            

def download_amr_reference_file() -> pd.DataFrame:
    url = 'https://ftp.ncbi.nlm.nih.gov/pathogen/Antimicrobial_resistance/AMRFinderPlus/database/latest/ReferenceGeneCatalog.txt'
    response = requests.get(url)
    df = pd.read_csv(io.BytesIO(response.content), sep='\t')
    df['element'] = np.where(df['allele'].notna(), df['allele'], df['gene_family'])
    df = df[['element', 'product_name', 'class', 'subclass', 'hierarchy_node']]
    df = df.drop_duplicates()
    return df


import pandas as pd

from typing import Any

from google.cloud import bigquery

def query_set_of_clusters(biosamples: list[str]) -> list[str]:
    biosamples_str = python_list_to_sql_str(biosamples)
    query = f'''
    SELECT DISTINCT erd_group AS cluster
    FROM `ncbi-pathogen-detect.pdbrowser.isolates`
    WHERE biosample_acc IN ({biosamples_str})
    '''
    df = execute_query(query)
    clusters = df['cluster'].to_list()
    return clusters

def query_isolates(clusters: list[str], biosamples: list[str]) -> pd.DataFrame:
    clusters_str = python_list_to_sql_str(clusters)
    biosamples_str = python_list_to_sql_str(biosamples)
    query = f'''
    SELECT
        isolate_identifiers[0] AS isolate_id,
        biosample_acc AS biosample,
        target_acc,
        erd_group AS cluster,
        Run AS sra_id,
        isolation_source,
        geo_loc_name,
        collection_date,
        creation_date,
        taxgroup_name,
        scientific_name,
        bioproject_acc,
        STRING_AGG(a.element, ",") AS amr_genotypes_elements,
        STRING_AGG(a.method, ",") AS amr_genotypes_methods
    FROM `ncbi-pathogen-detect.pdbrowser.isolates`
    LEFT JOIN UNNEST(AMR_genotypes) AS a
    WHERE erd_group IN ({clusters_str})
    OR biosample_acc IN ({biosamples_str}) 
    GROUP BY
        isolate_id,
        biosample,
        target_acc,
        cluster,
        sra_id,
        isolation_source,
        geo_loc_name,
        collection_date,
        creation_date,
        taxgroup_name,
        scientific_name,
        bioproject_acc
    ORDER BY isolate_id;
    '''
    df = execute_query(query)
    df['collection_date'] = df['collection_date'].astype('string')
    
    def interleave_amr_fields(row: pd.Series) -> str:
        result = ''
        if row['amr_genotypes_elements'] is None:
            return result
        elements = row['amr_genotypes_elements'].split(',')
        methods = row['amr_genotypes_methods'].split(',')
        for element, method in zip(elements, methods):
            result += f'{element}={method},'
        result = result.rstrip(',')
        return result

    df['amr_genotypes'] = df.apply(interleave_amr_fields, axis=1)
    df = df.drop(['amr_genotypes_elements', 'amr_genotypes_methods'], axis=1)
    return df


def query_clusters(biosamples: list[str]) -> pd.DataFrame:
    biosamples_str = python_list_to_sql_str(biosamples)
    query = f'''
    SELECT DISTINCT
        cluster_isolates.erd_group AS cluster,
        cluster_size.num AS total_count,
        taxgroup_name,
        earliest_added,
        latest_added,
        earliest_year_collected,
        latest_year_collected
    FROM
    (
        SELECT
            erd_group,
            taxgroup_name
        FROM `ncbi-pathogen-detect.pdbrowser.isolates`
        WHERE biosample_acc IN ({biosamples_str})
        GROUP BY
            erd_group,
            taxgroup_name
    ) AS cluster_isolates
    LEFT JOIN
    (
        SELECT
            erd_group,
            COUNT(*) AS num,
            MIN(SUBSTRING(creation_date, 0, 10)) AS earliest_added,
            MAX(SUBSTRING(creation_date, 0, 10)) AS latest_added,
            MIN(SUBSTRING(collection_date, 0, 4)) AS earliest_year_collected,
            MAX(SUBSTRING(collection_date, 0, 4)) as latest_year_collected
        FROM `ncbi-pathogen-detect.pdbrowser.isolates`
        GROUP BY erd_group
    ) AS cluster_size
    ON cluster_isolates.erd_group = cluster_size.erd_group
    WHERE cluster_size.num IS NOT NULL
    ORDER BY total_count DESC
    '''
    df = execute_query(query)
    return df

def python_list_to_sql_str(python_list: list[str]) -> str:
    quoted_list = [f'"{l}"' for l in python_list]
    sql_str = ', '.join(quoted_list)
    return sql_str

def execute_query(query: str) -> pd.DataFrame:
    client = bigquery.Client()
    df = client.query_and_wait(query).to_dataframe()
    return df

def isolates_df_from_browser_df(
    browser_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Parse DataFrame from --browser-tsv into `isolates_df` DataFrame format.
    The output format should match that from query_isolates().
    """
    browser_df['isolate_id'] = browser_df['Isolate identifiers'].str.split(',').str[0]
    rename_cols = {
        'isolate_id': 'isolate_id',
        'BioSample': 'biosample',
        'Isolate': 'target_acc',
        'SNP cluster': 'cluster',
        'Run': 'sra_id',
        'Isolation source': 'isolation_source',
        'Location': 'geo_loc_name',
        'Collection date': 'collection_date',
        'Create date': 'creation_date',
        '#Organism group': 'taxgroup_name',
        'Scientific name': 'scientific_name',
        'BioProject': 'bioproject_acc',
        'AMR genotypes': 'amr_genotypes',
    }
    missing_cols = [k for k in rename_cols if k not in browser_df.columns]
    if missing_cols:
        raise ValueError(
            f'Required columns {missing_cols} missing in browser file'
        )
    df = browser_df[rename_cols.keys()].rename(columns=rename_cols)
    df['collection_date'] = df['collection_date'].astype('string')
    return df 

def cluster_df_from_isolates_df(
    isolates_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    Return DataFrame with counts and date ranges for each cluster in
    `isolates_df`.
    """
    isolates_df = isolates_df.copy()
    cluster_isolates = (
        isolates_df
        .groupby(['cluster', 'taxgroup_name'], as_index=False)
        .first()
    )
    isolates_df['creation_date'] = pd.to_datetime(
        isolates_df['creation_date'], errors='coerce'
    )
    isolates_df['collection_date'] = isolates_df['collection_date'].astype('string')
    isolates_df['collection_year'] = isolates_df['collection_date'].str[:4]
    cluster_size = (
        isolates_df
        .groupby('cluster')
        .agg(
            total_count=('cluster', 'count'),
            earliest_added=(
                'creation_date',
                lambda x: x.dropna().min().strftime('%Y-%m-%d'),
            ),
            latest_added=(
                'creation_date',
                lambda x: x.dropna().max().strftime('%Y-%m-%d'),
            ),
            earliest_year_collected=(
                'collection_year',
                lambda x: x.dropna().min(),
            ),
            latest_year_collected=(
                'collection_year',
                lambda x: x.dropna().max(),
            )
        )
        .reset_index()
    )

    df = (
        cluster_isolates
        .merge(cluster_size, on='cluster', how='left')
        .dropna(subset=['total_count'])
        .sort_values(by='total_count', ascending=False)
    )
    keep_cols = [
        'cluster',
        'total_count',
        'taxgroup_name',
        'earliest_added',
        'latest_added',
        'earliest_year_collected',
        'latest_year_collected'
    ]
    df = df[keep_cols] 
    df = df.astype({
            'earliest_year_collected': 'string',
            'latest_year_collected': 'string',
        })
    return df


def create_amr_df(
    isolates_df: pd.DataFrame,
    amr_ref_df: pd.DataFrame
) -> pd.DataFrame:

    amr_df = isolates_df.copy()[['isolate_id', 'biosample', 'amr_genotypes']]
    amr_df['amr_genotype'] = amr_df['amr_genotypes'].str.split(',')
    amr_df = amr_df.explode('amr_genotype')
    amr_df['element'] = amr_df['amr_genotype'].str.split('=').str[0]
    amr_df['method'] = amr_df['amr_genotype'].str.split('=').str[1]
    amr_df = amr_df.merge(amr_ref_df, how='left', on=['element'])
    amr_df = amr_df[['isolate_id', 'biosample', 'element', 'method', 'product_name', 'class', 'subclass']]
    return amr_df


def filter_amr_df(amr_df: pd.DataFrame, filters: list[str]) -> pd.DataFrame:
    filtered_dfs = []
    for filter in filters:
        item = filter.split(':')
        _class = item[0].upper()
        subclass = item[1].upper()
        filtered_df = amr_df[
            (amr_df['class'].str.contains(_class))
            & (amr_df['subclass'].str.contains(subclass))
        ]
        filtered_dfs.append(filtered_df)
    concat_df = pd.concat(filtered_dfs)
    filtered_df = concat_df.drop_duplicates()
    return filtered_df


import argparse

from importlib.metadata import version

from typing import Sequence

def parse_args(command: Sequence[str]) -> argparse.Namespace:
    """
    Parse command-line arguments from the user.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        'sample_sheet',
        help='Path to sample sheet CSV with required "biosample" column and any additional metadata columns. Use "id" column for alternate isolate IDs.',
    )
    parser.add_argument(
        '--out-dir', '-o',
        help='Path to directory to store outputs. Defaults to "./outputs/" if not specified.'
    )
    parser.add_argument(
        '--retry',
        help='Do not query BigQuery or NCBI, assumes data has already been downloaded to --out-dir.',
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--browser-file',
        # TODO link to instructions
        help='Path to isolates TSV or CSV downloaded from the Pathogen Detection Isolates Browser with information for all internal and external isolates. When specified, data in file will be used instead of querying the BigQuery dataset.'
    )
    parser.add_argument(
        '--keep-snp-files',
        help='Keep downloaded SNP and tree files in the output directory. By default, files are deleted after processing.',
        action='store_true',
    )
    parser.add_argument(
        '--amr',
        help='Include AMR tab in report with antimicrobial resistance genes detected by AMRFinderPlus.',
        action='store_true',
    )
    parser.add_argument(
        '--filter-amr',
        help='Only include AMR genes in provided comma-separated list of CLASS:SUBCLASS pairs in the AMR tab. Also adds filtered_amr column to Isolates and Cluster details tab and matching genes to tree labels',
        type=lambda s: [i for i in s.split(',')],
    )
    parser.add_argument(
        '--version', '-v',
        help='Print the version of ncbi-cluster-tracker and exit.',
        action='version',
        version=version('ncbi-cluster-tracker'),
    )
    mutex_group_compare = parser.add_mutually_exclusive_group()
    mutex_group_compare.add_argument(
        '--compare-dir',
        help='Path to previous output directory to detect and report new isolates.',
    )
    args = parser.parse_args(command)

    if args.retry and not args.out_dir:
        parser.error('--retry flag requires --out_dir argument')

    if args.filter_amr:
        if not args.amr:
            parser.error('--filter-amr argument requires --amr flag')
        for item in args.filter_amr:
            if ':' not in item[1:-1]:
                parser.error('Each element in --filter-amr list must be in the form CLASS:SUBCLASS')
    
    return args




import glob
import os

import dendropy # type: ignore
import pandas as pd
import tqdm

import ncbi_cluster_tracker.cluster as cluster

from ncbi_cluster_tracker.logger import logger

class Cluster:
    """
    Represents a SNP cluster on NCBI Pathogen Detection and its distance matrix. 
    """
    MAX_MATRIX_SIZE = 20
    MAX_TREE_SIZE = 500

    def __init__(self, name: str, internal_isolates: list[str]):
        self.name = name
        self.cluster_id = name.split('.')[0]
        self.internal_isolates = set(internal_isolates)
        self.external_isolates: set[str] | None = None
        self.filtered_matrix_message: str | None = None
        self.filtered_matrix: pd.DataFrame | None = self._create_filtered_matrix()
    
    def _create_filtered_matrix(self) -> pd.DataFrame | None:
        """
        Generate distance matrix from downloaded Pathogen Detection tree,
        with number of isolates filtered down to a reasonably viewable number.
        """
        tree_file_glob = f'{self.cluster_id}*.newick_tree.newick'
        tree_path = glob.glob(
            os.path.join(os.environ['NCT_OUT_SUBDIR'], 'snps', tree_file_glob
        ))[0]
        tree = dendropy.Tree.get_from_path(tree_path, schema='newick')

        internal_taxa = {t for t in tree.taxon_namespace
                         if t.label in self.internal_isolates}

        external_taxa = {t for t in tree.taxon_namespace
                         if t.label not in self.internal_isolates}

        self.external_isolates = {t.label for t in external_taxa}

        # Arakawa warns if matrix has > 500 cells so limit matrix
        if len(tree) > self.MAX_MATRIX_SIZE:
            if len(tree.taxon_namespace) > self.MAX_TREE_SIZE:
                if 1 < len(internal_taxa) < self.MAX_MATRIX_SIZE:
                    # filter down to just the first self.MAX_TREE_SIZE taxa
                    tree.retain_taxa(internal_taxa)
                    self.filtered_matrix_message = ( 
                        f'SNP distance matrix filtered to show only internal isolates '
                        f'since there are more than {self.MAX_TREE_SIZE} isolates in the cluster.'
                    )
                elif len(internal_taxa) == 1:
                    tree = None
                    self.filtered_matrix_message = (
                        f'SNP distance matrix not displayed since there are '
                        f'more than {self.MAX_TREE_SIZE} isolates and only '
                        f'one internal isolate in the cluster.'
                    )

                else:
                    tree = None
                    self.filtered_matrix_message = (
                        f'SNP distance matrix not displayed since there are '
                        f'more than {self.MAX_TREE_SIZE} isolates in the cluster.'
                    )

            elif len(internal_taxa) == self.MAX_MATRIX_SIZE:
                # filter down taxa to just input isolates
                tree.retain_taxa(internal_taxa)
                self.filtered_matrix_message = ( 
                    f'SNP distance matrix filtered to show only the {self.MAX_MATRIX_SIZE} '
                    f'internal isolates.'
                )
            elif len(internal_taxa) < self.MAX_MATRIX_SIZE:
                # filter down to include only the nearest external isolates
                matrix = tree.phylogenetic_distance_matrix().as_data_table()
                matrix_df = pd.DataFrame.from_records(matrix._data)
                num_externals_to_keep = self.MAX_MATRIX_SIZE - len(internal_taxa)
                minimums = {}
                for external in self.external_isolates:
                    distances = matrix_df[external]
                    filtered_distances = distances.loc[
                        matrix_df.index.isin(self.internal_isolates)
                    ]
                    minimums[external] = filtered_distances.min() 
                sorted_externals = sorted(minimums.items(), key=lambda i: i[1])
                nearest_externals = [
                    k for k, _ in sorted_externals[:num_externals_to_keep]
                ]
                nearest_external_taxa = {
                    t for t in tree.taxon_namespace if t.label in nearest_externals
                }
                filtered_taxa = internal_taxa.union(nearest_external_taxa)
                tree.retain_taxa(filtered_taxa)
                self.filtered_matrix_message  = ( 
                    f'SNP distance matrix filtered to show only the '
                    f'the nearest {num_externals_to_keep} external isolate(s).'
                )
            else:
                tree = None
                self.filtered_matrix = None
                self.filtered_matrix_message = (
                    f'SNP distance matrix not displayed since there are over '
                    f'{self.MAX_MATRIX_SIZE} internal isolates in the cluster.'
                ) 
        if tree is not None:
            matrix = tree.phylogenetic_distance_matrix().as_data_table()
            matrix_df = pd.DataFrame.from_records(matrix._data)
            matrix_df.columns = matrix_df.columns.astype(str)
            matrix_df.index = matrix_df.index.astype(str)
            matrix_df = matrix_df.astype(int)

            # Sort matrix based on preorder tree traversal
            taxa = [n.taxon for n in tree if n.taxon is not None]
            labels = [n.label for n in taxa]
            matrix_df = matrix_df.reindex(index=labels, columns=labels)

            return matrix_df
        return None


def create_clusters(
        sample_sheet_df: pd.DataFrame,
        isolates_df: pd.DataFrame,
        clusters_df: pd.DataFrame,
    ) -> list[cluster.Cluster]:
    """
    Create list of all of the Clusters and their associated isolates.
    """
    clusters: list[cluster.Cluster] = []
    logger.info('Creating clusters...')
    for cluster_name in tqdm.tqdm(clusters_df['cluster'].tolist()):
        internal_isolates = isolates_df[
            (isolates_df['cluster'] == cluster_name)
            & (isolates_df['biosample'].isin(sample_sheet_df.index))
        ]['target_acc'].tolist()
        clusters.append(cluster.Cluster(cluster_name, internal_isolates))
    return clusters




import argparse
import datetime
import glob
import os
import re
import shutil
import sys

import arakawa as ar # type: ignore
import pandas as pd
import plotly.express as px # type: ignore

from ncbi_cluster_tracker import cli
from ncbi_cluster_tracker import cluster
from ncbi_cluster_tracker import download
from ncbi_cluster_tracker import query
from ncbi_cluster_tracker import report

from ncbi_cluster_tracker.logger import logger

def main() -> None:
    command = f'{os.path.basename(sys.argv[0])} {" ".join(sys.argv[1:])}'
    args = cli.parse_args(sys.argv[1:])
    sample_sheet_df = (pd
        .read_csv(args.sample_sheet, dtype={'id': 'string'})
        .set_index('biosample', verify_integrity=True)
    )
    biosamples = sample_sheet_df.index.to_list()

    out_dir = 'outputs' if args.out_dir is None else args.out_dir
    
    if args.compare_dir is None:
        old_clusters_df = None
        old_isolates_df = None
    else:
        old_clusters_glob = glob.glob(os.path.join(args.compare_dir, '*clusters*.csv'))
        old_isolates_glob = glob.glob(os.path.join(args.compare_dir, '*isolates*.csv'))
        if not old_clusters_glob:
            raise FileNotFoundError(f'Could not find clusters CSV file in {args.compare_dir}')
        if len(old_clusters_glob) > 1:
            raise ValueError(f'Multiple clusters CSV files found in {args.compare_dir}')
        if not old_isolates_glob:
            raise FileNotFoundError(f'Could not find isolates CSV file in {args.compare_dir}')
        if len(old_isolates_glob) > 1:
            raise ValueError(f'Multiple isolates CSV files found in {args.compare_dir}')
        old_clusters_df = pd.read_csv(old_clusters_glob[0])
        old_isolates_df = pd.read_csv(old_isolates_glob[0])

    if not args.retry:
        os.environ['NCT_NOW'] = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        os.environ['NCT_OUT_SUBDIR'] = os.path.join(out_dir, os.environ['NCT_NOW'])
        os.makedirs(os.environ['NCT_OUT_SUBDIR'], exist_ok=True)
        if args.browser_file is not None:
            isolates_df, clusters_df = get_clusters(biosamples, args.browser_file)
        else:
            isolates_df, clusters_df = get_clusters(biosamples, 'bigquery')
        download.download_cluster_files(clusters_df, args.keep_snp_files)
    else:
        if not os.path.isdir(out_dir):
            raise FileNotFoundError(f'Could not find existing output directory {out_dir} for --retry')
        os.environ['NCT_OUT_SUBDIR'] = out_dir
        os.environ['NCT_NOW'] = os.path.basename(out_dir.rstrip(os.sep))
        logger.info(f'Retrying with {os.environ["NCT_OUT_SUBDIR"]}')
        isolates_df, clusters_df = get_clusters(biosamples, 'local')
    
    if args.amr:
        amr_ref_df = download.download_amr_reference_file()  
        amr_df = query.create_amr_df(isolates_df, amr_ref_df)
        if args.filter_amr:
            amr_df = query.filter_amr_df(amr_df, args.filter_amr)
    else:
        amr_df = None

    clusters_df['tree_url'] = clusters_df.apply(download.build_tree_viewer_url, axis=1)
    clusters = cluster.create_clusters(sample_sheet_df, isolates_df, clusters_df)
    if not args.keep_snp_files:
        shutil.rmtree(os.path.join(os.environ['NCT_OUT_SUBDIR'], 'snps'))
    isolates_df = report.mark_new_isolates(isolates_df, old_isolates_df)
    metadata = report.combine_metadata(sample_sheet_df, isolates_df, amr_df, args)
    report.write_final_report(
        clusters_df,
        old_clusters_df,
        clusters,
        metadata,
        amr_df,
        args.compare_dir,
        command,
    )


def get_clusters(
    biosamples: list[str],
    data_location: str
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Fetch cluster data from NCBI's BigQuery `pdbrowser` dataset for the given
    `biosamples` if `data_location` is 'bigquery', or read from existing output
    CSV files if `data_location` is 'local`, otherwise read from --browser-tsv
    file.

    Return `isolates_df` DataFrame with isolate-level metadata, and
    `clusters_df` DataFrame with cluster-level metadata. Additionally, the
    DataFrames' data is written to a CSV in the output directory.
    """
    isolates_csv = os.path.join(
        os.environ['NCT_OUT_SUBDIR'],
        f'isolates_{os.environ["NCT_NOW"]}.csv'
    )
    clusters_csv = os.path.join(
        os.environ['NCT_OUT_SUBDIR'],
        f'clusters_{os.environ["NCT_NOW"]}.csv'
    )

    if data_location == 'bigquery':
        clusters = query.query_set_of_clusters(biosamples)
        isolates_df = query.query_isolates(clusters, biosamples)
        # TODO: query_clusters() should be replaceable with cluster_df_from_isolates_df()
        clusters_df = query.query_clusters(biosamples)
        isolates_df.to_csv(isolates_csv, index=False)
    elif data_location == 'local':
        try:
            isolates_df = pd.read_csv(isolates_csv)
            clusters_df = pd.read_csv(clusters_csv)
        except FileNotFoundError as e:
            message = f'Existing isolates/clusters CSV not found in {os.environ["NCT_OUT_SUBDIR"]}'
            raise Exception(message) from e
    else:
        if data_location.endswith('.tsv'):
            browser_df = pd.read_csv(data_location, sep='\t', low_memory=False, on_bad_lines='warn')
        elif data_location.endswith('.csv'):
            browser_df = pd.read_csv(data_location, low_memory=False, on_bad_lines='warn')
        else:
            raise ValueError(f'Invalid file type (must be .tsv or .csv): {data_location}')
        isolates_df = query.isolates_df_from_browser_df(browser_df)
        clusters_df = query.cluster_df_from_isolates_df(isolates_df)
        isolates_df.to_csv(isolates_csv, index=False)

    return isolates_df, clusters_df


if __name__ == '__main__':
    main()